from relationalai.early_access.lqp import ir as lqp
from relationalai.early_access.metamodel import ir

rel_to_lqp = {
    "+": "rel_primitive_add",
    "-": "rel_primitive_subtract",
    "*": "rel_primitive_multiply",
    "/": "rel_primitive_divide",
    "=": "rel_primitive_eq",
    "!=": "rel_primitive_neq",
    "<=": "rel_primitive_lt_eq",
    ">=": "rel_primitive_gt_eq",
    ">": "rel_primitive_gt",
    "<": "rel_primitive_lt",
    "construct_date": "rel_primitive_construct_date",
    "construct_datetime": "rel_primitive_construct_datetime",
}

def relname_to_lqp_name(name: str) -> str:
    # TODO: do these proprly
    if name in rel_to_lqp:
        return rel_to_lqp[name]
    else:
        raise NotImplementedError(f"missing primitive case: {name}")

def lqp_sum_op() -> lqp.Abstraction:
    # TODO: make sure gensym'd properly
    vs = [
        lqp.Var("x", lqp.PrimitiveType.INT),
        lqp.Var("y", lqp.PrimitiveType.INT),
        lqp.Var("z", lqp.PrimitiveType.INT),
    ]

    body = lqp.Primitive("rel_primitive_add", [vs[0], vs[1], vs[2]])
    return lqp.Abstraction(vs, body)

def lqp_operator(op: ir.Relation) -> lqp.Abstraction:
    if op.name == "sum":
        return lqp_sum_op()
    elif op.name == "count":
        return lqp_sum_op()
    else:
        raise NotImplementedError(f"Unsupported aggregation: {op.name}")



from relationalai.early_access.metamodel.util import NameCache
from relationalai.early_access.lqp.validators import assert_valid_input, assert_valid_update
from relationalai.early_access.metamodel import ir, builtins as rel_builtins, helpers
from relationalai.early_access.lqp import ir as lqp
from relationalai.early_access.lqp.hash_utils import lqp_hash
from relationalai.early_access.lqp.primitives import relname_to_lqp_name, lqp_operator
from relationalai.early_access.lqp.types import meta_type_to_lqp, type_from_constant, type_from_term
from relationalai.early_access.lqp.constructors import mk_and, mk_exists, mk_or
import datetime as dt

from typing import Tuple, cast

class TranslationCtx:
    def __init__(self, model):
        # TODO: comment these fields
        # TODO: should we have a pass to rename variables instead of this?
        self.var_name_cache = NameCache()
        self.id_to_orig_name = {}
        self.output_ids = []

""" Main access point. Converts the model IR to an LQP program. """
def to_lqp(model: ir.Model) -> lqp.LqpProgram:
    assert_valid_input(model)
    ctx = TranslationCtx(model)
    program = _translate_to_program(ctx, model)
    return program

def _translate_to_program(ctx: TranslationCtx, model: ir.Model) -> lqp.LqpProgram:
    decls: list[lqp.Declaration] = []
    outputs: list[Tuple[str, lqp.RelationId]] = []

    # LQP only accepts logical tasks
    # These are asserted at init time
    root = cast(ir.Logical, model.root)

    seen_rids = set()
    for subtask in root.body:
        assert isinstance(subtask, ir.Logical)
        decl = _translate_to_decl(ctx, subtask)

        assert isinstance(decl, lqp.Def), "we dont do loops yet m8"
        decls.append(decl)

        rid = decl.name
        assert rid not in seen_rids, f"duplicate relation id: {rid}"
        seen_rids.add(rid)

    for output_id in ctx.output_ids:
        assert isinstance(output_id, lqp.RelationId)
        outputs.append(("output", output_id))

    debug_info = mk_debug_info(ctx)

    return lqp.LqpProgram(decls, outputs, debug_info)

def mk_debug_info(ctx: TranslationCtx) -> lqp.DebugInfo:
    return lqp.DebugInfo(ctx.id_to_orig_name)

def _make_def(ctx: TranslationCtx, name: str, projection: list[lqp.Var], conjs: list[lqp.Formula], exist_vars: list[lqp.Var], is_output: bool = False) -> lqp.Def:
    abstraction = lqp.Abstraction(
        projection,
        mk_exists(
            exist_vars,
            mk_and(conjs)
        ),
    )
    rel_id = name_to_id(name)
    ctx.id_to_orig_name[rel_id] = name

    if is_output:
        ctx.output_ids.append(rel_id)

    # TODO: is this correct? might need attrs tooo?
    return lqp.Def(rel_id, abstraction, [])

# Split the exists into the existential vars, the aggregates, and the other tasks
def _split_exists(ctx: TranslationCtx, task: ir.Exists) -> Tuple[list[ir.Aggregate], list[ir.Task]]:
    if isinstance(task.task, ir.Aggregate):
        return [task.task], []

    if not isinstance(task.task, ir.Logical):
        return [], []

    aggrs = []
    other_tasks = []
    for subtask in task.task.body:
        if isinstance(subtask, ir.Aggregate):
            aggrs.append(subtask)
        else:
            other_tasks.append(subtask)

    return aggrs, other_tasks

def _translate_to_decl(ctx: TranslationCtx, rule: ir.Logical) -> lqp.Declaration:
    effects = []
    body_tasks = []
    aggregates = []

    exist_ir_vars = []

    for task in rule.body:
        if isinstance(task, ir.Output) or isinstance(task, ir.Update):
            effects.append(task)
        elif isinstance(task, ir.Lookup):
            body_tasks.append(task)
        elif isinstance(task, ir.Logical):
            body_tasks.append(task)
        elif isinstance(task, ir.Exists):
            aggrs, others = _split_exists(ctx, task)
            exist_ir_vars.extend(task.vars)
            aggregates.extend(aggrs)
            body_tasks.extend(others)
        elif isinstance(task, ir.Not):
            body_tasks.append(task)
        elif isinstance(task, ir.Aggregate):
            aggregates.append(task)
        elif isinstance(task, ir.Construct):
            body_tasks.append(task)
        elif isinstance(task, ir.Union):
            body_tasks.append(task)
        else:
            raise NotImplementedError(f"Unknown task type: {type(task)}")

    conjuncts = [_translate_to_formula(ctx, task) for task in body_tasks]

    exist_vars, eqs = translate_bindings(ctx, exist_ir_vars)
    conjuncts.extend(eqs)

    assert len(effects) == 1, f"should only have exactly one effect, got {len(effects)}"
    effect = effects[0]

    if isinstance(effect, ir.Output):
        assert len(aggregates) == 0, "cannot have aggregates in output"
        is_output = True
        def_name = "output"
        projection = []

        # TODO: we dont yet handle aliases, so we ignore v[0]
        bindings = [v[1] for v in effect.aliases]
        projection, eqs = translate_bindings(ctx, bindings)
        conjuncts.extend(eqs)

        return _make_def(ctx, def_name, projection, conjuncts, exist_vars, is_output)

    assert isinstance(effect, ir.Update), f"expected update, got {type(effect)}"
    assert_valid_update(effect)
    def_name = effect.relation.name
    is_output = False

    # Handle the bindings
    projection, eqs = translate_bindings(ctx, list(effect.args))
    conjuncts.extend(eqs)

    # Aggregates reduce over the body
    if len(aggregates) > 0:
        aggr_body = mk_and(conjuncts)
        conjuncts = []
        for aggr in aggregates:
            # TODO: what to do with the other result?
            reduce, _ = _translate_aggregate(ctx, aggr, aggr_body, effect)
            conjuncts.append(reduce)
    return _make_def(ctx, def_name, projection, conjuncts, exist_vars, is_output)

def _translate_aggregate(ctx: TranslationCtx, aggr: ir.Aggregate, body: lqp.Formula, update: ir.Update) -> Tuple[lqp.Reduce, list[lqp.Var]]:
    # TODO: handle this properly
    aggr_name = aggr.aggregation.name
    assert aggr_name == "sum" or aggr_name == "count", f"only support sum or count for now, not {aggr.aggregation.name}"

    group_bys = [_translate_term(ctx, var) for var in aggr.group]
    projected_args = [_translate_term(ctx, var) for var in aggr.projection]

    # TODO: differentiate between results and more args
    result_args = []

    # TODO: not sure if this si right
    translated_vars = []
    for arg in aggr.args:
        assert isinstance(arg, ir.Var)
        translated_vars.append(_translate_term(ctx, arg))

    # TODO: input and output should be checked using the aggr not like this
    # Last one is output arg
    output_var = translated_vars[-1]
    result_args.append(output_var)

    # The rest are input args
    input_args = []
    for arg in translated_vars[:-1]:
        input_args.append(arg)

    inner_abstr_args = []
    inner_abstr_args.extend(projected_args)
    inner_abstr_args.extend(input_args)
    if aggr_name == "count":
        # Count sums up "1"
        one_var, eq = binding_to_lqp_var(ctx, 1)
        assert eq is not None
        body = mk_and([body, eq])
        inner_abstr_args.append(one_var)

    outer_abstr_args = []
    for arg in group_bys:
        outer_abstr_args.append(arg)
    outer_abstr_args.append(output_var)

    op = lqp_operator(aggr.aggregation)
    inner_abstr = lqp.Abstraction(
        inner_abstr_args,
        body,
    )
    reduce = lqp.Reduce(
        op,
        inner_abstr,
        result_args,
    )
    return reduce, outer_abstr_args

def _translate_to_formula(ctx: TranslationCtx, task: ir.Task) -> lqp.Formula:
    if isinstance(task, ir.Logical):
        conjuncts = [_translate_to_formula(ctx, child) for child in task.body]
        return mk_and(conjuncts)
    elif isinstance(task, ir.Lookup):
        return _translate_to_atom(ctx, task)
    elif isinstance(task, ir.Not):
        return lqp.Not(_translate_to_formula(ctx, task.task))
    elif isinstance(task, ir.Exists):
        lqp_vars, conjuncts = translate_bindings(ctx, list(task.vars))
        conjuncts.append(_translate_to_formula(ctx, task.task))
        return mk_exists(lqp_vars, mk_and(conjuncts))
    elif isinstance(task, ir.Construct):
        assert len(task.values) >= 1, "construct should have at least one value"
        # TODO: what does the first value do
        terms = [_translate_term(ctx, arg) for arg in task.values[1:]]
        terms.append(_translate_term(ctx, task.id_var))
        return lqp.Primitive(
            "rel_primitive_hash_tuple_uint128",
            terms,
        )
    elif isinstance(task, ir.Union):
        # TODO: handle hoisted vars if needed
        assert len(task.hoisted) == 0, "hoisted updates not supported yet, because idk what it means"
        disjs = [_translate_to_formula(ctx, child) for child in task.tasks]
        return mk_or(disjs)
    else:
        raise NotImplementedError(f"Unknown task type (formula): {type(task)}")

def _translate_term(ctx: TranslationCtx, value: ir.Value) -> lqp.Term:
    if isinstance(value, ir.Var):
        name = ctx.var_name_cache.get_name(value.id, value.name)
        t = meta_type_to_lqp(value.type)
        return lqp.Var(name, t)
    elif isinstance(value, ir.Literal):
        return lqp.Constant(value.value)
    elif isinstance(value, str):
        return lqp.Constant(value)
    elif isinstance(value, int):
        return lqp.Constant(value)
    elif isinstance(value, dt.date):
        return lqp.Constant(value)
    else:
        raise NotImplementedError(f"Unknown value type: {type(value)}")

def _translate_to_atom(ctx: TranslationCtx, task: ir.Lookup) -> lqp.Formula:
    # TODO: want signature not name
    rel_name = task.relation.name
    terms = []
    sig_types = []
    for arg in task.args:
        if isinstance(arg, lqp.PrimitiveValue):
            term = lqp.Constant(arg)
            terms.append(term)
            t = type_from_constant(arg)
            sig_types.append(t)
            continue
        elif isinstance(arg, ir.Literal):
            term = lqp.Constant(arg.value)
            terms.append(term)
            t = type_from_constant(arg.value)
            sig_types.append(t)
            continue
        assert isinstance(arg, ir.Var), f"expected var, got {type(arg)}: {arg}"
        var = _translate_term(ctx, arg)
        terms.append(var)
        sig_types.append(type_from_term(var))

    # TODO: wrong
    if rel_builtins.is_builtin(task.relation):
        lqp_name = relname_to_lqp_name(task.relation.name)
        return lqp.Primitive(lqp_name, terms)

    if helpers.is_external(task.relation):
        return lqp.RelAtom(
            task.relation.name,
            terms,
        )

    rid = get_relation_id(ctx, rel_name, sig_types)
    return lqp.Atom(rid, terms)

def get_relation_id(ctx: TranslationCtx, name: str, types: list[lqp.PrimitiveType]) -> lqp.RelationId:
    relation_id = name_to_id(name)
    ctx.id_to_orig_name[relation_id] = name
    return relation_id

# TODO: should this take types too?
def name_to_id(name: str) -> lqp.RelationId:
    return lqp.RelationId(lqp_hash(name))

def translate_bindings(ctx: TranslationCtx, bindings: list[ir.Value]) -> Tuple[list[lqp.Var], list[lqp.Formula]]:
    lqp_vars = []
    conjuncts = []
    for binding in bindings:
        lqp_var, eq = binding_to_lqp_var(ctx, binding)
        lqp_vars.append(lqp_var)
        if eq is not None:
            conjuncts.append(eq)

    return lqp_vars, conjuncts

def binding_to_lqp_var(ctx: TranslationCtx, binding: ir.Value) -> Tuple[lqp.Var, lqp.Union[None, lqp.Formula]]:
    if isinstance(binding, ir.Var):
        lqp_var = _translate_term(ctx, binding)
        assert isinstance(lqp_var, lqp.Var)
        return lqp_var, None
    else:
        # Constant in this case
        assert isinstance(binding, lqp.PrimitiveValue)
        lqp_value = _translate_term(ctx, binding)

        # TODO: gensym
        var_name = ctx.var_name_cache.get_name(1, "cvar")
        typ = type_from_constant(binding)

        lqp_var = lqp.Var(var_name, typ)
        eq = lqp.Primitive("rel_primitive_eq", [lqp_var, lqp_value])
        return lqp_var, eq



from relationalai.early_access.lqp import ir as lqp
from typing import Sequence
import hashlib


# TODO: this is NOT a good hash its just to get things working for now to get
# a stable id.

"""
Hashes an LQP node to a 128-bit integer.
"""
def lqp_hash(node) -> int:
    if isinstance(node, lqp.LqpNode):
        return hash_to_uint128(_lqp_hash(node))
    else:
        h = _lqp_hash_fn(node)
        return hash_to_uint128(h)

def _lqp_hash(node: lqp.LqpNode) -> int:
    if isinstance(node, lqp.Abstraction):
        h1 = _lqp_hash_list(node.vars)
        h2 = _lqp_hash(node.value)
        return _lqp_hash_fn((h1, h2))
    elif isinstance(node, lqp.Conjunction):
        h1 = _lqp_hash_list(node.args)
        return _lqp_hash_fn((h1,))
    elif isinstance(node, lqp.Var):
        return _lqp_hash_fn((node.name, node.type))
    elif isinstance(node, lqp.Constant):
        return _lqp_hash_fn((node.value,))
    elif isinstance(node, lqp.Primitive):
        h1 = _lqp_hash_fn(node.name)
        h2 = _lqp_hash_list(node.terms)
        return _lqp_hash_fn((h1, h2))
    elif isinstance(node, lqp.Reduce):
        h1 = _lqp_hash(node.op)
        h2 = _lqp_hash(node.body)
        h3 = _lqp_hash_list(node.terms)
        return _lqp_hash_fn((h1, h2, h3))
    elif isinstance(node, lqp.Atom):
        h1 = _lqp_hash(node.name)
        h2 = _lqp_hash_list(node.terms)
        return _lqp_hash_fn((h1, h2))
    elif isinstance(node, lqp.RelationId):
        return _lqp_hash_fn((node.id,))
    else:
        raise NotImplementedError(f"Unsupported LQP node type: {type(node)}")

# TODO: this is NOT a good hash its just to get things working for now to get
# a stable id.
def _lqp_hash_fn(node) -> int:
    return int.from_bytes(hashlib.sha256(str(node).encode()).digest(), byteorder='big', signed=False)

def _lqp_hash_list(node: Sequence[lqp.LqpNode]) -> int:
    hashes = [_lqp_hash(n) for n in node]
    return _lqp_hash_fn(tuple(hashes))

def hash_to_uint128(h: int) -> int:
    return h % (2**128)  # Ensure it's within the 128-bit range



from relationalai.early_access.metamodel import ir as meta
from relationalai.early_access.metamodel import types
from relationalai.early_access.lqp import ir as lqp
import datetime

def type_from_term(arg: lqp.Term) -> lqp.RelType:
    if isinstance(arg, lqp.Var):
        return arg.type
    elif isinstance(arg, lqp.Constant):
        return type_from_constant(arg.value)
    else:
        raise NotImplementedError(f"Unknown term type: {type(arg)}")

def meta_type_to_lqp(type: meta.Type) -> lqp.RelType:
    if isinstance(type, meta.UnionType):
        # TODO - this is WRONG! we need to fix the typer wrt overloading
        type = type.types.some()

    assert isinstance(type, meta.ScalarType)
    if types.is_builtin(type):
        # TODO: just ocompare to types.py
        if type.name == "Int":
            return lqp.PrimitiveType.INT
        elif type.name == "Float":
            return lqp.PrimitiveType.FLOAT
        elif type.name == "String":
            return lqp.PrimitiveType.STRING
        elif type.name == "Number":
            # TODO: fix this, this is wrong
            return lqp.PrimitiveType.INT
        elif type.name == "Decimal":
            return lqp.ValueType.DECIMAL
        elif type.name == "Date":
            return lqp.ValueType.DATE
        elif type.name == "DateTime":
            return lqp.ValueType.DATETIME
        elif type.name == "Any":
            # TODO: fix this, this is wrong
            return lqp.PrimitiveType.UNKNOWN
    elif (types.is_entity_type(type)):
        return lqp.PrimitiveType.UINT128
    raise NotImplementedError(f"Unknown type: {type.name}")

def type_from_constant(arg: lqp.PrimitiveValue) -> lqp.RelType:
    if isinstance(arg, int):
        return lqp.PrimitiveType.INT
    elif isinstance(arg, float):
        return lqp.PrimitiveType.FLOAT
    elif isinstance(arg, str):
        return lqp.PrimitiveType.STRING
    # TODO: Direct use of date/datetime is not supported in the IR, so this should be
    # rewritten with construct_date.
    elif isinstance(arg, datetime.date):
        return lqp.ValueType.DATE
    else:
        raise NotImplementedError(f"Unknown constant type: {type(arg)}")



from __future__ import annotations
from typing import cast, Optional

from relationalai.early_access.metamodel import ir, factory as f, helpers, visitor
from relationalai.early_access.metamodel.compiler import Pass, group_tasks
from relationalai.early_access.metamodel.util import OrderedSet, ordered_set
from relationalai.early_access.rel import metamodel_utils
from relationalai.early_access.rel.dependency import AnalysisContext, ctx_frame

class Flatten(Pass):
    """
    Traverses the model's root to flatten it as much as possible. The result of this pass is
    a Logical root where all nested tasks that represent a rule in Rel are extraced to the
    top level.

    - nested logical with updates becomes a top-level logical (a rule)

    From:
        Logical
            Logical
                lookup1   <- scope is spread
                Logical
                    lookup2
                    derive foo
                Logical
                    lookup3
                    derive bar
    To:
        Logical
            Logical
                lookup1
                lookup2
                derive foo
            Logical
                lookup1
                lookup3
                derive bar

    - nested logical with aggregates becomes a top-level logical (a rule representing an aggregation)

    From:
        Logical
            Logical
                lookup1
                Logical
                    lookup2
                    aggregate1
                Logical
                    lookup3
                    aggregate2
                output
    To:
        Logical
            Logical
                lookup1
                lookup2
                aggregate1
                derive tmp1
            Logical
                lookup1
                lookup3
                aggregate2
                derive tmp2
            Logical
                lookup1
                lookup tmp1
                lookup tmp2
                output

    - a union becomes a top-level logical for each branch, writing into a temporary relation,
    and a lookup from that relation.

    From:
        Logical
            Logical
                Union
                    Logical
                        lookup1
                    Logical
                        lookup2
                output
    To:
        Logical
            Logical
                lookup1
                derive tmp1
            Logical
                lookup2
                derive tmp1
            Logical
                lookup tmp1
                output

    - a match becomes a top-level logical for each branch, each writing into its own temporary
    relation and a lookup from the last relation. The top-level logical for a branch derives
    into the temporary relation negating the previous branch:

    From:
        Logical
            Logical
                Match
                    Logical
                        lookup1
                    Logical
                        lookup2
                output
    To:
        Logical
            Logical
                lookup1
                derive tmp1
            Logical
                Union            <- tmp1() or (not temp1() and lookup2())
                    lookup tmp1
                    Logical
                        Not
                            lookup tmp1
                        lookup2
                        derive tmp2
            Logical
                lookup tmp2
                output
    """

    #--------------------------------------------------
    # Public API
    #--------------------------------------------------
    def rewrite(self, model: ir.Model, cache) -> ir.Model:
        # create the dependency analysis context
        ctx = AnalysisContext(model.root)

        # rewrite the root
        replacement = self.handle(model.root, ctx)

        # the new body contains the extracted top level logicals and maybe the rewritten root
        body = ctx.top_level if replacement is None else ctx.top_level + [replacement]

        # create the new model, updating relations and root
        return ir.Model(
            model.engines,
            OrderedSet.from_iterable(model.relations).update(ctx.relations).frozen(),
            model.types,
            ir.Logical(model.root.engine, tuple(), tuple(body))
        )

    #--------------------------------------------------
    # IR handlers
    #--------------------------------------------------

    def handle(self, task: ir.Task, ctx: AnalysisContext) -> Optional[ir.Task]:
        if isinstance(task, ir.Logical):
            return self.handle_logical(task, ctx)
        elif isinstance(task, ir.Union):
            return self.handle_union(task, ctx)
        elif isinstance(task, ir.Match):
            return self.handle_match(task, ctx)
        elif isinstance(task, ir.Require):
            return self.handle_require(task, ctx)
        else:
            return task

    def handle_logical(self, task: ir.Logical, ctx: AnalysisContext):
        # create frame to process the children
        with ctx_frame(ctx, task) as frame:

            # keep track of what's the result of handling nested composites
            composites = group_tasks(task.body, {
                "composites": helpers.COMPOSITES
            })["composites"]
            all_composites_removed = len(composites) > 0

            # recursively handle children, collecting the replacements in the body
            body:OrderedSet[ir.Task] = ordered_set()
            for child in task.body:
                replacement = self.handle(child, ctx)
                if replacement is not None:
                    frame.replaced(child, replacement)
                    extend_body(body, replacement)
                    # nested composite was not completely removed
                    if child in composites:
                        all_composites_removed = False

            # all children were extracted or all composites were removed without any outputs
            # (so no way for outer dependencies), drop this logical
            if not body or (all_composites_removed and not frame.task_outputs(task)):
                return None

            # now process the rewritten body
            groups = group_tasks(body.list, {
                "outputs": ir.Output,
                "updates": ir.Update,
                "aggregates": ir.Aggregate,
                "ranks": ir.Rank,
            })

            # if there are outputs, currently assume it's already at top level, so just return
            # the rewritten body
            if groups["outputs"]:
                return ir.Logical(task.engine, task.hoisted, tuple(body), task.annotations)

            # if there are updates, extract as a new top level rule
            if groups["updates"]:
                # add dependencies for the logical, which is in the parent frame
                if frame._parent_frame:
                    body = frame._parent_frame.task_dependencies(task) | body
                ctx.top_level.append(ir.Logical(task.engine, task.hoisted, tuple(body), task.annotations))

                # no need to refer to the extracted logical because it is an update
                return None

            if groups["aggregates"]:
                if len(groups["aggregates"]) > 1:
                    # stop rewritting as we don't know how to handle this yet
                    return task

                # there must be only one
                agg = cast(ir.Aggregate, groups["aggregates"].some())

                # add agg dependencies to the body
                body = frame.task_dependencies(agg) | body

                # extract a new logical for the aggregate, exposing aggregate group-by and results
                exposed_vars = OrderedSet.from_iterable(list(agg.group) + helpers.aggregate_outputs(agg))
                connection = metamodel_utils.extract(agg, body, exposed_vars.list, ctx)

                # return a reference to the connection relation
                reference = f.logical([f.lookup(connection, exposed_vars.list)], merge_var_list(exposed_vars.list, task.hoisted))
                return reference

            if groups["ranks"]:
                if len(groups["ranks"]) > 1:
                    # stop rewritting as we don't know how to handle this yet
                    return task

                # there must be only one
                rank = cast(ir.Rank, groups["ranks"].some())

                # add rank dependencies to the body
                body = frame.task_dependencies(rank) | body

                # extract a new logical for the rank, exposing rank group-by and results
                exposed_vars = OrderedSet.from_iterable(list(rank.group) + list(rank.args) + [rank.result])
                connection = metamodel_utils.extract(rank, body, exposed_vars.list, ctx)

                # return a reference to the connection relation
                reference = f.logical([f.lookup(connection, exposed_vars.list)], merge_var_list(exposed_vars.list, task.hoisted))
                return reference

            return ir.Logical(task.engine, task.hoisted, tuple(body))

    def handle_match(self, match: ir.Match, ctx: AnalysisContext):
        # TODO: how to deal with malformed input like this?
        if not match.tasks:
            return match

        frame = ctx.current_frame()
        body = frame.task_dependencies(match)
        outputs = frame.task_outputs(match)
        exposed_vars = (frame.task_inputs(match) | outputs).list

        connection = None
        reference = None

        for branch in match.tasks:
            # process the branch
            replacement = self.handle(branch, ctx)
            assert(replacement)
            frame.replaced(branch, replacement)

            branch_body: OrderedSet[ir.Task] = OrderedSet.from_iterable(body)
            extend_body(branch_body, replacement)
            if reference:
                branch_body.add(negate(reference, len(outputs)))
                branch_body = OrderedSet.from_iterable([f.union([f.logical(branch_body.list, match.hoisted), reference], match.hoisted)])
            connection = metamodel_utils.extract(branch, branch_body, exposed_vars, ctx, "match")
            reference = f.logical([f.lookup(connection, exposed_vars)], exposed_vars)

        return reference


    def handle_union(self, union: ir.Union, ctx: AnalysisContext):
        # TODO: how to deal with malformed input like this?
        if not union.tasks:
            return union

        frame = ctx.current_frame()
        body = frame.task_dependencies(union)
        outputs = frame.task_outputs(union)
        exposed_vars = (frame.task_inputs(union) | outputs).list

        connection = None

        for branch in union.tasks:
            # process the branch
            replacement = self.handle(branch, ctx)
            assert(replacement)
            frame.replaced(branch, replacement)

            branch_body: OrderedSet[ir.Task] = OrderedSet.from_iterable(body)
            extend_body(branch_body, replacement)

            if connection is None:
                # first branch, extract making a connection relation
                connection = metamodel_utils.extract(branch, branch_body, exposed_vars, ctx, "union")
            else:
                # subsequent branch, extract reusing the connection relation
                # add derivation to the extracted body
                branch_body.add(f.derive(connection, exposed_vars))

                # extract the body
                ctx.top_level.append(ir.Logical(union.engine, tuple(), tuple(branch_body)))

        # return a reference to the connection
        assert(connection)
        reference = f.logical([f.lookup(connection, exposed_vars)], exposed_vars)
        return reference

    def handle_require(self, req: ir.Require, ctx: AnalysisContext):
        # only extract the domain if it is a somewhat complex Logical and there's more than
        # one check, otherwise insert it straight into all checks
        domain = req.domain
        if len(req.checks) > 1 and isinstance(domain, ir.Logical) and len(domain.body) > 1:
            body = OrderedSet.from_iterable(domain.body)
            vars = helpers.hoisted_vars(domain.hoisted)
            connection = metamodel_utils.extract(req, body, vars, ctx, "domain")
            domain = f.logical([f.lookup(connection, vars)], vars)

        for check in req.checks:
            # only generate logic for checks that have errors
            if check.error:
                handled_check = self.handle(check.check, ctx)
                if handled_check:
                    body = ordered_set()
                    body.add(domain)
                    body.add(ir.Not(req.engine, handled_check))
                    if (isinstance(check.error, ir.Logical)):
                        body.update(check.error.body)
                    else:
                        # this is more general but may trip the current splinter
                        body.add(check.error)
                    ctx.top_level.append(ir.Logical(req.engine, tuple(), tuple(body)))

        # currently we just drop the Require, but we should keep it here and link the
        # extracted logicals to it
        return None


#--------------------------------------------------
# Helpers
#--------------------------------------------------

def extractable(t: ir.Task):
    """
    Whether this task is a Logical that will be extracted as a top level by this
    pass, because it has an aggregation, effects, match, union, etc.
    """
    extractable_types = (ir.Update, ir.Aggregate, ir.Match, ir.Union, ir.Rank)
    return isinstance(t, ir.Logical) and len(visitor.collect_by_type(extractable_types, t)) > 0

def extractables(composites: OrderedSet[ir.Task]):
    """ Filter the set of composites, keeping only the extractable ones. """
    return list(filter(extractable, composites))

def negate(reference: ir.Logical, values: int):
    """
    Return a negation of this reference, where the last `values` arguments are to
    be replaced by wildcards (i.e. len(reference.args) - values are keys so they need
    to be bound in the Not.)
    """
    lookup = cast(ir.Lookup, reference.body[0])
    args = []
    i = 0
    last = len(lookup.args) - values
    for arg in lookup.args:
        args.append(f.wild()) if i >= last else args.append(arg)
        i += 1

    return ir.Not(reference.engine, f.lookup(lookup.relation, args))

def merge_var_list(vars: list[ir.Var], hoisted: tuple[ir.VarOrDefault, ...]) -> list[ir.VarOrDefault]:
    """ Merge vars and hoisted, making sure that hoisted vars have precedence since they may have defaults. """
    r = []
    hoisted_vars = helpers.hoisted_vars(hoisted)
    for v in vars:
        if v not in hoisted_vars:
            r.append(v)
    r.extend(hoisted)
    return r

def extend_body(body: OrderedSet[ir.Task], extra: ir.Task):
    """ Add the extra task to the body, but if the extra is a simple logical, just
    inline its subtasks. """
    if isinstance(extra, ir.Logical):
        if extra.hoisted:
            # hoists, remove things that are already in the body to avoid duplicates
            logical_body = []
            for t in extra.body:
                if t not in body:
                    logical_body.append(t)
            if len(logical_body) == len(extra.body):
                # no duplicates
                body.add(extra)
            else:
                # some duplicate, remove them
                body.add(ir.Logical(
                    extra.engine,
                    extra.hoisted,
                    tuple(logical_body)
                ))
        else:
            # no hoists, just inline
            body.update(extra.body)
    else:
        body.add(extra)



from .cdc import CDC
from .extract_common import ExtractCommon
from .flatten import Flatten
from .quantify_vars import QuantifyVars

__all__ = [
    "CDC",
    "ExtractCommon",
    "Flatten",
    "QuantifyVars",
]



from typing import Any, Optional

from relationalai.early_access.builder import Concept as QBConcept


class Concept(QBConcept):

    def __init__(self, model, name: str, extends: list[Any] = []):
        super().__init__(name, extends, model.qb_model())
        self._dsl_model = model

    def _is_value_type(self) -> bool:
        if len(self._extends) == 1:
            ext_concept = self._extends[0]
            if ext_concept._is_primitive():
                return True
            else:
                return ext_concept._is_value_type()
        return False

    def __eq__(self, other):
        return isinstance(other, Concept) and self._name == other._name

    def __hash__(self):
        return hash(self._name)


class EntityType(Concept):

    def __init__(self, model, nm, extends: list[Concept] = [], ref_schema_name: Optional[str] = None):
        self._domain = extends
        super().__init__(model, nm, extends)
        self.__ref_schema_name = ref_schema_name

    def _qualified_name(self):
        return self._name

    def _is_composite(self):
        return len(self._domain) > 1

    def _ref_schema_name(self):
        return self.__ref_schema_name

    def _is_value_type(self) -> bool:
        return False


class ValueType(Concept):

    def __init__(self, model, nm, extends: Optional[Concept] = None):
        super().__init__(model, nm, [extends] if extends is not None else [])

    def _is_value_type(self) -> bool:
        return True


from __future__ import annotations
from dataclasses import dataclass
from enum import Enum
from typing import Union, Tuple, Sequence
import datetime as dt

# Tree representation of LQP. Each non-terminal (those with more than one
# option) is an "abstract" class and each terminal is its own class. All of
# which are children of LqpNode. PrimitiveType and PrimitiveValue are
# exceptions. PrimitiveType is an enum and PrimitiveValue is just a value.
# https://docs.google.com/document/d/1QXRU7zc1SUvYkyMCG0KZINZtFgzWsl9-XHxMssdXZzg/

@dataclass(frozen=True)
class LqpNode:
    pass

@dataclass(frozen=True)
class DebugInfo:
    id_to_orig_name: dict[RelationId, str]

# Declaration := Def | Loop
@dataclass(frozen=True)
class Declaration(LqpNode):
    pass

@dataclass(frozen=True)
class LqpProgram:
    defs: list[Declaration]
    # name -> relation id
    outputs: list[Tuple[str, RelationId]]
    # optional debug info
    debug_info: Union[DebugInfo,None]=None

# Def(name::RelationId, body::Abstraction, attrs::Attribute[])
@dataclass(frozen=True)
class Def(Declaration):
    name: RelationId
    body: Abstraction
    attrs: Sequence[Attribute]

# Loop(temporal_var::LoopIndex, inits::Def[], body::Declaration[])
@dataclass(frozen=True)
class Loop(Declaration):
    temporal_var: str
    inits: Sequence[Def]
    body: Declaration

# Abstraction := Abstraction(vars::Var[], value::Formula)
@dataclass(frozen=True)
class Abstraction(LqpNode):
    vars: Sequence[Var]
    value: Formula

# Formula := Exists
#          | Reduce
#          | Conjunction
#          | Disjunction
#          | Not
#          | FFI
#          | Atom
#          | Pragma
#          | Primitive
#          | True
#          | False
#          | RelAtom
@dataclass(frozen=True)
class Formula(LqpNode):
    pass

# Exists(vars::Var[], value::Formula)
@dataclass(frozen=True)
class Exists(Formula):
    body: Abstraction

# Reduce(op::Abstraction, body::Abstraction, terms::Term[])
@dataclass(frozen=True)
class Reduce(Formula):
    op: Abstraction
    body: Abstraction
    terms: Sequence[Term]

# Conjunction(args::Formula[])
@dataclass(frozen=True)
class Conjunction(Formula):
    args: Sequence[Formula]

# Disjunction(args::Formula[])
@dataclass(frozen=True)
class Disjunction(Formula):
    args: Sequence[Formula]

# Not(arg::Formula)
@dataclass(frozen=True)
class Not(Formula):
    arg: Formula

# FFI(name::Symbol, args::Abstraction[], terms::Term[])
@dataclass(frozen=True)
class Ffi(Formula):
    name: str
    args: Sequence[Abstraction]
    terms: Sequence[Term]

# Atom(name::RelationId, terms::Term[])
@dataclass(frozen=True)
class Atom(Formula):
    name: RelationId
    terms: Sequence[Term]

# Pragma(name::Symbol, terms::Term[])
@dataclass(frozen=True)
class Pragma(Formula):
    name: str
    terms: Sequence[Term]

# Primitive(name::Symbol, terms::Term[])
@dataclass(frozen=True)
class Primitive(Formula):
    name: str
    terms: Sequence[Term]

# RelAtom(name::String, terms::Term[])
@dataclass(frozen=True)
class RelAtom(Formula):
    name: str
    terms: Sequence[Term]

# Term := Var | Constant
@dataclass(frozen=True)
class Term(LqpNode):
    pass

# Var(name::Symbol, type::PrimitiveType)
@dataclass(frozen=True)
class Var(Term):
    name: str
    type: RelType

# Constant(value::PrimitiveValue)
@dataclass(frozen=True)
class Constant(Term):
    value: PrimitiveValue

# Attribute := Attribute(name::Symbol, args::Constant[])
@dataclass(frozen=True)
class Attribute(LqpNode):
    name: str
    args: Sequence[Constant]

# RelationId := RelationId(id::UInt128)
@dataclass(frozen=True)
class RelationId(LqpNode):
    # We use a catchall int here to represent the uint128 as it is difficult
    # to do so in Python without external packages. We check the value in
    # __post_init__.
    id: int

    def __post_init__(self):
        if self.id < 0 or self.id > 0xffffffffffffffffffffffffffffffff:
            raise ValueError(
                "RelationId constructed with out of range (UInt128) number: {}"
                    .format(self.id)
            )

# ValueType
class ValueType(Enum):
    # TODO: fill this out properly
    UNKNOWN = 0
    DECIMAL = 1
    DATE = 2
    DATETIME = 3
    NANOSECOND = 4
    MICROSECOND = 5
    MILLISECOND = 6
    SECOND = 7
    MINUTE = 8
    HOUR = 9
    DAY = 10
    WEEK = 11
    MONTH = 12
    YEAR = 13

# PrimitiveType := STRING | DECIMAL | INT | FLOAT | HASH
# TODO: we don't know what types we're supporting yet.
class PrimitiveType(Enum):
    # TODO: get rid of this oen maybe?
    UNKNOWN = 0
    STRING = 1
    INT = 2
    FLOAT = 3
    UINT128 = 4

# RelType := PrimitiveType | ValueType
RelType = Union[PrimitiveType, ValueType]

# PrimitiveValue := string | decimal | int | float | hash
# TODO: we don't know what types we're supporting yet.
PrimitiveValue = Union[str, int, float, dt.date]



from collections import OrderedDict
from typing import Optional, Any, List, Union, TypeVar

from relationalai import Config
from relationalai.early_access.builder import Model as QBModel
from relationalai.early_access.builder import Relationship as QBRelationship
from relationalai.early_access.dsl.bindings.tables import Binding
from relationalai.early_access.dsl.bindings.tables import SnowflakeTable, CsvTable
from relationalai.early_access.dsl.codegen.weaver import Weaver
from relationalai.early_access.dsl.core.types.unconstrained import UnconstrainedValueType
from relationalai.early_access.dsl.core.utils import generate_stable_uuid, to_pascal_case
from relationalai.early_access.dsl.orm.constraints import Unique, Mandatory, RoleValueConstraint, Range, \
    ExclusiveSubtypeConstraint, InclusiveSubtypeConstraint
from relationalai.early_access.dsl.orm.relationships import Relationship, Relation, Role
from relationalai.early_access.dsl.orm.types import Concept, EntityType, ValueType
from relationalai.early_access.dsl.snow.api import Executor
from relationalai.early_access.metamodel.util import OrderedSet

T = TypeVar('T', int, float, str)

class Model:
    def __init__(
            self,
            name: str,
            is_primary: bool = True,
            dry_run: bool = False,
            config: Optional[Config] = None
    ):
        self.name = name
        self._qb_model = QBModel(name, dry_run=dry_run, config=config)
        self.is_primary = is_primary
        self._value_types = OrderedDict()
        self._entity_types = OrderedDict()
        self._relationships = OrderedDict()
        self._relations = OrderedDict()
        self._constraints = OrderedSet()
        self._entity_to_id_relationship = OrderedDict()
        self._bindable_tables = OrderedDict()
        self._bindings = []
        self._executor = None
        self._weaver = Weaver(self)
        self._queries = OrderedSet()

    def guid(self):
        return generate_stable_uuid(self.name)

    def ValueType(self, name: str, extends: Optional['Concept']=None) -> ValueType:
        name = to_pascal_case(name)
        vt = ValueType(self, name, extends)
        self._validate_type_name(name)
        self._value_types[name] = vt
        if name not in self.qb_model().concepts:
            self.qb_model().concepts[name] = vt
        return vt

    def EntityType(self, name: str, extends:list['Concept'] = [], refmode: Optional[str] = None) -> EntityType:
        name = to_pascal_case(name)
        et = EntityType(self, name, extends, ref_schema_name=refmode)
        self._validate_type_name(name)
        self._entity_types[name] = et
        if name not in self.qb_model().concepts:
            self.qb_model().concepts[name] = et
        return et

    def Concept(self, name: str, extends: list['Concept'] = []) -> 'Concept':
        name = to_pascal_case(name)
        c = Concept(self, name, extends)
        if name not in self.qb_model().concepts:
            self.qb_model().concepts[name] = c
        self._validate_type_name(name)
        if c._is_value_type():
            self._value_types[name] = c
        else:
            self._entity_types[name] = c
        return c

    def qb_model(self):
        return self._qb_model

    def Relationship(self, reading: Any, relation_name: Optional[str] = None) -> QBRelationship:
        """
        Create a new relationship with the given reading.
        """
        relationship = Relationship(self, reading, relation_name=relation_name)
        self._add_relationship(relationship)
        return relationship._readings_map[reading]

    def constraint(self, constraint):
        # Generic constraint addition with validation
        self._constraints.add(constraint)
        constraint._desugar()

    def unique(self, *roles):
        self.constraint(Unique(*roles))

    def mandatory(self, role):
        self.constraint(Mandatory(role))

    def role_value_constraint(self, role: Role, values: List[Union[T, Range[T]]]):
        first_type = values[0]._type() if isinstance(values[0], Range) else type(values[0])
        for v in values[1:]:
            t = v._type() if isinstance(v, Range) else type(v)
            if t != first_type:
                raise Exception("Values for role value constraint must have the same type.")
        if role.player()._is_primitive() or role.player()._is_value_type():
            self.constraint(RoleValueConstraint(role, values))
        else:
            raise Exception("A role value constraint can only be applied to roles played by value types")

    def exclusive_subtype_constraint(self, *concepts:'Concept'):
        self.constraint(ExclusiveSubtypeConstraint(*concepts))

    def inclusive_subtype_constraint(self, *concepts:'Concept'):
        self.constraint(InclusiveSubtypeConstraint(*concepts))

    def ref_scheme(self, *relations:Relation):
        self._validate_ref_scheme_input_relations(relations)
        if len(relations) == 1:
            # binary case, internal UC
            role = relations[0][1]
            self._internal_preferred_uc(role)
        else:
            roles = [rel[1] for rel in relations]
            self._composite_preferred_uc(*roles)

    def _internal_preferred_uc(self, role):
        rel = role._part_of()
        if rel._arity() != 2:
            raise Exception("The relationship should be binary to apply preferred identifier constraint")
        # mark the role as preferred identifier
        self.constraint(Unique(role, is_preferred_identifier=True))
        # mark the sibling role as mandatory and unique
        sibling = role.sibling()
        self.mandatory(sibling)
        self.unique(sibling)

    def _composite_preferred_uc(self, *roles):
        for role in roles:
            sibling = role.sibling()
            if not sibling:
                raise Exception("Composite preferred identifier constraint should be applied on binary relationships")
            self.mandatory(sibling)
            self.unique(sibling)
        self.constraint(Unique(*roles, is_preferred_identifier=True))

    @staticmethod
    def _validate_ref_scheme_input_relations(relations):
        if not relations:
            raise ValueError("ref_scheme requires at least one relation")
        ref_rel = relations[0]
        for rel in relations:
            if not isinstance(rel, QBRelationship):
                raise ValueError("ref_scheme can only be applied on Relation instances")
            if rel._arity() != 2:
                raise ValueError("ref_scheme can only be applied on binary relations")
            if rel._first().player()._id != ref_rel._first().player()._id:
                raise ValueError("For ref_scheme all relations must have the same role player in position 0")

    def lookup_concept(self, name) -> Optional['Concept']:
        name = to_pascal_case(name)
        if name in self._value_types:
            return self._value_types[name]
        if name in self._entity_types:
            return self._entity_types[name]
        return None

    def lookup_relationship(self, name) -> Optional['Relationship']:
        if name in self._relationships:
            return self._relationships[name]
        return None

    def constraints(self):
        """Getter for the _constraints property"""
        return self._constraints

    def queries(self):
        return self._queries

    def value_types(self):
        return list(self._value_types.values())

    def entity_types(self):
        return list(self._entity_types.values())

    def entity_types_map(self):
        return self._entity_types

    def relationships(self):
        return list(self._relationships.values())

    def relations(self):
        return list(self._relations.values())

    def bindable_tables(self):
        return self._bindable_tables

    def api(self):
        self._executor = self._executor or Executor(self._qb_model._config)
        return self._executor

    def table(self, name: str) -> SnowflakeTable:
        metadata = self.api().table_metadata(name)
        table = SnowflakeTable(metadata, self)
        self._bindable_tables[name] = table
        return table

    def csv_table(self, name: str, schema: dict[str, UnconstrainedValueType]) -> CsvTable:
        table = CsvTable(name, schema, self)
        self._bindable_tables[name] = table
        return table

    def binding(self, binding: Binding):
        self._bindings.append(binding)

    def generate_weaving_rules(self):
        self._weaver.generate()

    def _add_relationship(self, rel: 'Relationship'):
        self._validate_relationship_name(rel._name)
        self._relationships[rel._name] = rel

    def _add_relation(self, relation: Relation) -> Relation:
        if relation._id in self._relations:
            existed_relation = self._relations[relation._id]
            # todo: add validation that relations' signature is the same
            return existed_relation
        self._relations[relation._id] = relation
        return relation

    def _validate_type_name(self, name):
        if name in self._entity_types:
            raise Exception(
                f"The name '{name}' is used to declare an entity type.")
        if name in self._value_types:
            raise Exception(
                f"The name '{name}' is used to declare a value type.")

    def _validate_relationship_name(self, name):
        if name in self._relationships:
            raise Exception(
                f"The name '{name}' is used to declare a relationship.")



from __future__ import annotations
from typing import Any, Union

from .. import builder as b
from . import dates, math, strings

def _make_expr(op: str, *args: Any) -> b.Expression:
    return b.Expression(b.Relationship.builtins[op], *args)

def range(*args: Union[b.Producer, int]) -> b.Expression:
    # supports range(stop), range(start, stop), range(start, stop, step)
    if len(args) == 1:
        start, stop, step = 0, args[0], 1
    elif len(args) == 2:
        start, stop, step = args[0], args[1], 1
    elif len(args) == 3:
        start, stop, step = args
    else:
        raise ValueError(f"range expects 1, 2, or 3 arguments, got {len(args)}")
    # unlike Python, Rel's range is 1..stop inclusive, so we need to subtract 1 from stop
    return _make_expr("range", start, stop-1, step, b.Integer.ref("res"))

__all__ = [
    "range",
    "dates",
    "math",
    "strings",
]



from typing import Union

from . import ir

from relationalai.lqp.v1 import logic_pb2, fragments_pb2, transactions_pb2

# pyright complains that *_pb2.Whatever is not a known attribute, so
# extract the classes with getattr.
__pb_Transaction = getattr(transactions_pb2, "Transaction")
__pb_Define = getattr(transactions_pb2, "Define")
__pb_Write = getattr(transactions_pb2, "Write")
__pb_Output = getattr(transactions_pb2, "Output")
__pb_Read = getattr(transactions_pb2, "Read")
__pb_Epoch = getattr(transactions_pb2, "Epoch")
__pb_Read = getattr(transactions_pb2, "Read")

__pb_Fragment = getattr(fragments_pb2, "Fragment")
__pb_FragmentId = getattr(fragments_pb2, "FragmentId")

__pb_Declaration = getattr(logic_pb2, "Declaration")
__pb_Def = getattr(logic_pb2, "Def")
__pb_Loop = getattr(logic_pb2, "Loop")
__pb_LoopIndex = getattr(logic_pb2, "LoopIndex")
__pb_Attribute = getattr(logic_pb2, "Attribute")
__pb_Abstraction = getattr(logic_pb2, "Abstraction")
__pb_Term = getattr(logic_pb2, "Term")
__pb_RelTerm = getattr(logic_pb2, "RelTerm")
__pb_Var = getattr(logic_pb2, "Var")
__pb_Constant = getattr(logic_pb2, "Constant")
__pb_Formula = getattr(logic_pb2, "Formula")
__pb_Exists = getattr(logic_pb2, "Exists")
__pb_Reduce = getattr(logic_pb2, "Reduce")
__pb_Conjunction = getattr(logic_pb2, "Conjunction")
__pb_Disjunction = getattr(logic_pb2, "Disjunction")
__pb_Not = getattr(logic_pb2, "Not")
__pb_FFI = getattr(logic_pb2, "FFI")
__pb_Atom = getattr(logic_pb2, "Atom")
__pb_RelAtom = getattr(logic_pb2, "RelAtom")
__pb_Pragma = getattr(logic_pb2, "Pragma")
__pb_Primitive = getattr(logic_pb2, "Primitive")
__pb_RelationId = getattr(logic_pb2, "RelationId")
__pb_PrimitiveValue = getattr(logic_pb2, "PrimitiveValue")
__pb_PrimitiveType = getattr(logic_pb2, "PrimitiveType")
__pb_RelType = getattr(logic_pb2, "RelType")
__pb_SpecializedValue = getattr(logic_pb2, "SpecializedValue")
__pb_RelValueType = getattr(logic_pb2, "RelValueType")

# Converts an LQP program to an LQP proto Transaction.
def lqp_program_to_proto(program: ir.LqpProgram):
    return __pb_Transaction(
        epochs = [
            __pb_Epoch(
                local_writes = [
                    __pb_Write(
                        define = __pb_Define(
                            fragment = __pb_Fragment(
                                id = __pb_FragmentId(id = bytes(404)),
                                declarations = [
                                    __as_declaration(lqp_node_to_proto(d)) for d in program.defs
                                ],
                            )
                        )
                    )
                ],
                reads = [
                    __pb_Read(
                        output = __pb_Output(
                            name = name,
                            relation_id = lqp_node_to_proto(rid),
                        )
                    ) for (name, rid) in program.outputs
                ],
            )
        ],
    )

TermOption = Union[__pb_Var, __pb_Constant]
# Creates a proto Term out of termish (an option for Term).
def __as_term(termish: TermOption):
    pb_term = __pb_Term()
    if isinstance(termish, __pb_Var):
        pb_term.var.CopyFrom(termish)
    else:
        assert isinstance(termish, __pb_Constant)
        pb_term.constant.CopyFrom(termish)

    return pb_term

RelTermOption = Union[__pb_Var, __pb_Constant, __pb_SpecializedValue]
def __as_rel_term(termish: RelTermOption):
    pb_term = __pb_RelTerm()
    if isinstance(termish, __pb_SpecializedValue):
        pb_term.specialized_value.CopyFrom(termish)
    else:
        assert isinstance(termish, TermOption)
        pb_term.term.CopyFrom(__as_term(termish))

    return pb_term

FormulaOption = Union[
    __pb_Exists, __pb_Reduce, __pb_Conjunction, __pb_Disjunction, __pb_Not,
    __pb_FFI, __pb_Atom, __pb_Pragma, __pb_Primitive,
]
# Creates a proto Formula out of formulaish (an option for Formula).
def __as_formula(formulaish: FormulaOption):
    pb_formula = __pb_Formula()
    if isinstance(formulaish, __pb_Exists):
        pb_formula.exists.CopyFrom(formulaish)
    elif isinstance(formulaish, __pb_Reduce):
        pb_formula.reduce.CopyFrom(formulaish)
    elif isinstance(formulaish, __pb_Conjunction):
        pb_formula.conjunction.CopyFrom(formulaish)
    elif isinstance(formulaish, __pb_Disjunction):
        pb_formula.disjunction.CopyFrom(formulaish)
    elif isinstance(formulaish, __pb_Not):
        getattr(pb_formula, "not").CopyFrom(formulaish)
    elif isinstance(formulaish, __pb_FFI):
        pb_formula.ffi.CopyFrom(formulaish)
    elif isinstance(formulaish, __pb_Atom):
        pb_formula.atom.CopyFrom(formulaish)
    elif isinstance(formulaish, __pb_RelAtom):
        pb_formula.rel_atom.CopyFrom(formulaish)
    elif isinstance(formulaish, __pb_Pragma):
        pb_formula.pragma.CopyFrom(formulaish)
    else:
        assert isinstance(formulaish, __pb_Primitive)
        pb_formula.primitive.CopyFrom(formulaish)

    return pb_formula

DeclarationOption = Union[__pb_Def, __pb_Loop]
# Creates a proto Declaration out of declarationish (an option for Declaration).
def __as_declaration(declarationish: DeclarationOption):
    pb_declaration = __pb_Declaration()
    if isinstance(declarationish, __pb_Def):
        getattr(pb_declaration, "def").CopyFrom(declarationish)
    else:
        assert isinstance(declarationish, __pb_Loop)
        pb_declaration.loop.CopyFrom(declarationish)

    return pb_declaration

# Produce proto out of LQP node recursively.
# TODO: should PrimitiveType be an LqpNode?
def lqp_node_to_proto(node: Union[ir.LqpNode, ir.PrimitiveType]):
    if isinstance(node, ir.Def):
        return __pb_Def(
            name = lqp_node_to_proto(node.name),
            body = lqp_node_to_proto(node.body),
            attrs = [lqp_node_to_proto(attr) for attr in node.attrs]
        )

    # elif isinstance(node, Declaration):
    #     Ignore: we match on the "concrete" children of Declaration.

    elif isinstance(node, ir.Loop):
        return __pb_Loop(
            temporal_var = __pb_LoopIndex(node.temporal_var),
            inits = [lqp_node_to_proto(init) for init in node.inits],
            body = lqp_node_to_proto(node.body),
        )

    elif isinstance(node, ir.Abstraction):
        return __pb_Abstraction(
            vars = [lqp_node_to_proto(var) for var in node.vars],
            value = __as_formula(lqp_node_to_proto(node.value)),
        )

    # elif isinstance(node, ir.Formula):
    #     Ignore: we match on the "concrete" children of Formula.

    elif isinstance(node, ir.Exists):
        return __pb_Exists(
            body = lqp_node_to_proto(node.body)
        )

    elif isinstance(node, ir.Reduce):
        return __pb_Reduce(
            op = lqp_node_to_proto(node.op),
            body = lqp_node_to_proto(node.body),
            terms = [__as_term(lqp_node_to_proto(term)) for term in node.terms],
        )

    elif isinstance(node, ir.Conjunction):
        return __pb_Conjunction(
            args = [__as_formula(lqp_node_to_proto(arg)) for arg in node.args],
        )

    elif isinstance(node, ir.Disjunction):
        return __pb_Disjunction(
            args = [__as_formula(lqp_node_to_proto(arg)) for arg in node.args],
        )

    elif isinstance(node, ir.Not):
        return __pb_Not(
            arg = __as_formula(lqp_node_to_proto(node.arg))
        )

    elif isinstance(node, ir.Ffi):
        return __pb_FFI(
            name = node.name,
            args = [lqp_node_to_proto(arg) for arg in node.args],
            terms = [__as_term(lqp_node_to_proto(term)) for term in node.terms]
        )

    elif isinstance(node, ir.Atom):
        return __pb_Atom(
            name = lqp_node_to_proto(node.name),
            terms = [__as_term(lqp_node_to_proto(term)) for term in node.terms],
        )

    elif isinstance(node, ir.Pragma):
        return __pb_Pragma(
            name = node.name,
            terms = [__as_term(lqp_node_to_proto(term)) for term in node.terms],
        )

    elif isinstance(node, ir.Primitive):
        return __pb_Primitive(
            name = node.name,
            terms = [__as_rel_term(lqp_node_to_proto(term)) for term in node.terms],
        )

    elif isinstance(node, ir.RelAtom):
        return __pb_RelAtom(
            name = node.name,
            terms = [__as_rel_term(lqp_node_to_proto(term)) for term in node.terms],
        )

    elif isinstance(node, ir.Var):
        return __pb_Var(
            name = node.name,
            type = lqp_type_to_proto(node.type),
        )

    elif isinstance(node, ir.Constant):
        # To put inside the constant.
        pb_primitive_value = __pb_PrimitiveValue()
        if isinstance(node.value, str):
            pb_primitive_value.string_value = node.value
        elif isinstance(node.value, int):
            pb_primitive_value.int_value = node.value
        elif isinstance(node.value, float):
            pb_primitive_value.float_value = node.value
        else:
            raise NotImplementedError(
                f"Proto transformation not implemented for {type(node.value)}."
            )
        return __pb_Constant(value = pb_primitive_value)

    elif isinstance(node, ir.Attribute):
        return __pb_Attribute(
            name = node.name,
            args = [lqp_node_to_proto(arg) for arg in node.args],
        )

    elif isinstance(node, ir.RelationId):
        return __pb_RelationId(
            id_low = 0x0000000000000000ffffffffffffffff & node.id,
            id_high = node.id >> 64,
        )

    else:
        raise NotImplementedError(
            f"Proto transformation not implemented for {type(node)}."
        )

def lqp_primitive_to_proto(node: ir.PrimitiveType):
    if node == ir.PrimitiveType.STRING:
        return __pb_PrimitiveType.PRIMITIVE_TYPE_STRING
    elif node == ir.PrimitiveType.INT:
        return __pb_PrimitiveType.PRIMITIVE_TYPE_INT
    elif node == ir.PrimitiveType.FLOAT:
        return __pb_PrimitiveType.PRIMITIVE_TYPE_FLOAT
    elif node == ir.PrimitiveType.UINT128:
        return __pb_PrimitiveType.PRIMITIVE_TYPE_UINT128
    elif node == ir.PrimitiveType.UNKNOWN:
        # TODO: this is wrong
        return __pb_PrimitiveType.PRIMITIVE_TYPE_INT
    else:
        raise NotImplementedError(
            f"Proto transformation not implemented for {node}."
        )

def lqp_value_type_to_proto(node: ir.ValueType):
    if node == ir.ValueType.DATE:
        return __pb_RelValueType.REL_VALUE_TYPE_DATE
    elif node == ir.ValueType.DATETIME:
        return __pb_RelValueType.REL_VALUE_TYPE_DATETIME
    else:
        raise NotImplementedError(
            f"Proto transformation not implemented for {node}."
        )

def lqp_type_to_proto(node: ir.RelType):
    if isinstance(node, ir.PrimitiveType):
        return __pb_RelType(
            primitive_type = lqp_primitive_to_proto(node)
        )

    elif isinstance(node, ir.ValueType):
        return __pb_RelType(
            value_type = lqp_value_type_to_proto(node)
        )

    else:
        raise NotImplementedError(
            f"Proto transformation not implemented for {type(node)}."
        )



"""
Type inference for the IR.
"""
from dataclasses import dataclass, field
from typing import Dict, Optional, List, Iterable, Union as PyUnion, Tuple, cast
from datetime import date, datetime
import networkx as nx

from relationalai.early_access.metamodel.util import OrderedSet, ordered_set
from relationalai.early_access.metamodel import ir, types, visitor, compiler

# Type System
# -----------
#
# Scalar types are either value types or entity types, not both.
# Value types (transitively) extend the built-in scalar types for which
# `is_value_base_type` holds.
# All other scalar types are entity types.
#
# Subtyping of scalar is declared explicitly in the model, currently as the
# `super_types` field in `ScalarType`, but this may change.
#
# The type system supports unions and tuples of types.
# Tuple types are covariant in their element types.
# Union types support the usual subtyping relation (for instance, a member of a
# union is a subtype of the union).
# We could, but currently do not, introduce common supertypes for unions.
#
# The `Any` type is used as a placeholder to mean "infer this type". `Any` is a
# scalar type.
# In particular, `Any` is not implicitly a supertype of any other type.
#
# List types are not allowed in the model and should have been rewritten to
# tuple types by the `RewriteListTypes` pass.
#
# Set types in the model indicate the multiplicity of the given field. They are
# not strictly types.
# For inference purposes, the element type is used as the type of the field,
# however when substituting inference results back into the model, the inferred
# type is wrapped in a `SetType`.
# Set types must be sets of scalar type (including `Any`).
#
# Type inference works as follows:
# - For each application of a relation:
#   - for input fields, bound the argument type by the field type (that is, arg
#     <: field; this is the standard rule for argument passing, types flow from
#     the argument to the field)
#   - for non-input fields (which could be input or output), equate the argument
#     type with the corresponding field type in the relation (= is more
#     efficient than adding two bounds: arg <: field and field <: arg)
# - For each variable, bound its type with its declared type.
# - For each field, bound its type with its declared type.
# - For default values, bound the variable type with the type of the default value.
# - These constraints build a graph of equivalence classes of nodes (fields and
# - vars). An edge a->b indicates that a is a subtype of b.
#   Each nodes has an associated set of upper and lower bounding scalar types.
# - Propagate upper bounds through the graph from supertype to subtype.
# - Propagate lower bounds through the graph from suptype to supertype.
# - Collapse cycles (SCCs) in the graph into an equivalence class.
# - For each equivalence class, union the upper bounds of all types in the
#   class. This is the inferred upper bound. Check that the inferred upper bound
#   of a node is a supertype of the all lower bounds of the node.
# - If the type of a field cannot be inferred to any type (as happens when the
#   field is not used in any application), equate the relation with other
#   relations of the same name and arity, if any. Then recompute the bound
#   again. This has the effect of treating relations of the same name and arity
#   as overloads of each other, but only when necessary.
# - Replace the types in the model with the intersection of the inferred upper
#   bound and the declared type. This strictly lowers types to be more precise.

@dataclass
class TypeVar:
    """A type variable."""
    # The node that this type variable is bound to.
    # Together with the index, this uniquely identifies the type variable.
    node: ir.Node = field(init=True)

    # The index of the field of the node, or 0.
    # Together with the node (id), this uniquely identifies the type variable.
    index: int = field(init=True)

    # The upper bound of the type variable.
    upper_bound: OrderedSet[ir.Type] = field(default_factory=OrderedSet, init=False, compare=False, hash=False, repr=False)
    # The lower bound of the type variable.
    lower_bound: OrderedSet[ir.Type] = field(default_factory=OrderedSet, init=False, compare=False, hash=False, repr=False)
    # Set of TypeVars that are supertypes of this type variable.
    node_super_types: OrderedSet["TypeVar"] = field(default_factory=OrderedSet, init=False, compare=False, hash=False, repr=False)

    # Union-find data structure.
    # The rank of the type variable.
    rank: int = field(default=0, init=False, compare=False, hash=False, repr=False)
    # The next type variable in the union-find data structure. If this is None, the type variable is the root of its union-find tree.
    next: Optional["TypeVar"] = field(default=None, init=False, compare=False, hash=False, repr=False)

    def __hash__(self) -> int:
        return hash((self.node.id, self.index))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TypeVar):
            return False
        return self.node.id == other.node.id and self.index == other.index

    def __str__(self) -> str:
        return f"{ir.node_to_string(self.node).strip()}{f'@{self.index}' if self.index >= 0 else ''} (id={self.node.id})"

    def find(self) -> "TypeVar":
        """Find the root of this type variable and perform path compression."""
        if self.next is None:
            return self
        self.next = self.next.find()
        return self.next

    def union(self, other: "TypeVar") -> "TypeVar":
        """Union this type variable with another, returning the root of the union."""
        top = self.find()
        bot = other.find()
        assert top.next is None
        assert bot.next is None
        if top == bot:
            return top
        if top.rank < bot.rank:
            top, bot = bot, top
        bot.next = top
        top.upper_bound.update(bot.upper_bound)
        top.lower_bound.update(bot.lower_bound)
        top.node_super_types.update(bot.node_super_types)
        bot.upper_bound.clear()
        bot.lower_bound.clear()
        bot.node_super_types.clear()
        assert self.find() == top
        assert other.find() == top
        return top

    @staticmethod
    def pretty_name(node: PyUnion[ir.Var, ir.Field], index: int, parent: Optional[ir.Node]) -> str:
        if isinstance(node, ir.Var):
            return f"Variable `{node.name}`"
        else:
            if index < 0:
                if isinstance(parent, ir.Relation):
                    return f"Field `{node.name}` of relation `{parent.name}`"
                else:
                    return f"Field `{node.name}`"
            else:
                if isinstance(parent, ir.Relation):
                    return f"Element {index} of field `{node.name}` of relation `{parent.name}`"
                else:
                    return f"Element {index} of field `{node.name}`"

@dataclass
class TypeError:
    """A type inference error."""
    msg: str
    node: ir.Node

Typeable = PyUnion[ir.Value, ir.Type, ir.Var, ir.Field, TypeVar]

@dataclass
class TypeEnv:
    """Environment for type inference that tracks type bounds for each node."""

    # The model being type-checked.
    model: ir.Model

    # Diagnostics. For now, this is just strings.
    diags: List[TypeError]

    # Maps node ids (and tuple index, as needed) to the type variables for their types.
    type_vars: Dict[Tuple[int, int], TypeVar]

    # How verbose to be with debug info, 0 for off.
    verbosity: int

    # Should we perform stricter checks on the inferred types.
    strict: bool

    def __init__(self, model: ir.Model, strict: bool, verbosity: int=0):
        super().__init__()
        self.model = model
        self.diags = []
        self.type_vars = {}
        self.strict = strict
        self.verbosity = verbosity

    def _complain(self, node: ir.Node, msg: str):
        """Report an error."""
        self.diags.append(TypeError(msg, node))

    def get_type_var(self, node: ir.Node, index: int = -1) -> TypeVar:
        key = (node.id, index)
        if key not in self.type_vars:
            self.type_vars[key] = TypeVar(node, index)
        tv = self.type_vars[key]
        assert tv.node == node
        assert tv.index == index
        return tv

    def _get_type(self, t: PyUnion[ir.Literal, ir.PyValue]) -> Optional[ir.Type]:
        if t is None:
            return types.Null
        elif isinstance(t, str):
            return types.String
        elif isinstance(t, bool):
            return types.Bool
        elif isinstance(t, int):
            return types.Int
        elif isinstance(t, float):
            return types.Float
        elif isinstance(t, date):
            return types.Date
        elif isinstance(t, datetime):
            return types.DateTime
        elif isinstance(t, ir.Literal):
            return t.type
        else:
            raise ValueError(f"Unexpected value {t} of type {type(t)}")

    def add_bound(self, lower: Typeable, upper: Typeable):
        if upper is None or isinstance(upper, (ir.Literal, str, bool, int, float, date, datetime)):
            t = self._get_type(upper)
            assert t is not None
            self.add_bound(lower, t)
        elif lower is None or isinstance(lower, (ir.Literal, str, bool, int, float, date, datetime)):
            t = self._get_type(lower)
            assert t is not None
            self.add_bound(t, upper)
        elif isinstance(lower, (ir.Var, ir.Field)):
            x = self.get_type_var(lower)
            self.add_bound(x, upper)
        elif isinstance(upper, (ir.Var, ir.Field)):
            x = self.get_type_var(upper)
            self.add_bound(lower, x)
        elif isinstance(upper, ir.UnionType):
            for t in upper.types:
                self.add_bound(lower, t)
        elif isinstance(lower, ir.UnionType):
            for t in lower.types:
                self.add_bound(t, upper)
        elif isinstance(lower, TypeVar):
            if isinstance(upper, ir.Type):
                if not types.is_any(upper):
                    lower.find().upper_bound.add(upper)
            elif isinstance(upper, TypeVar):
                lower.find().node_super_types.add(upper.find())
            else:
                raise ValueError(f"Unexpected types {type(lower)} and {type(upper)}")
        elif isinstance(lower, ir.Type):
            if isinstance(upper, ir.Type):
                if lower == upper:
                    return
                elif types.is_subtype(lower, upper):
                    return
                else:
                    self.diags.append(TypeError(f"Type {ir.type_to_string(lower)} is not a subtype of {ir.type_to_string(upper)}", lower))
            elif isinstance(upper, TypeVar):
                if not types.is_null(lower):
                    upper.find().lower_bound.add(lower)
            else:
                raise ValueError(f"Unexpected types {type(lower)} and {type(upper)}")
        else:
            raise ValueError(f"Unexpected types {type(lower)} and {type(upper)}")

    def add_equality(self, t1: Typeable, t2: Typeable):
        if t1 is None:
            self.add_equality(types.Null, t2)
        elif t2 is None:
            self.add_equality(t1, types.Null)
        elif isinstance(t1, (ir.Literal, str, bool, int, float, date, datetime)):
            t = self._get_type(t1)
            self.add_equality(t, t2)
        elif isinstance(t2, (ir.Literal, str, bool, int, float, date, datetime)):
            t = self._get_type(t2)
            self.add_equality(t1, t)
        elif isinstance(t1, (ir.Var, ir.Field)):
            x = self.get_type_var(t1)
            self.add_equality(x, t2)
        elif isinstance(t2, (ir.Var, ir.Field)):
            x = self.get_type_var(t2)
            self.add_equality(t1, x)
        elif isinstance(t1, ir.Type):
            if isinstance(t2, ir.Type):
                self.add_bound(t1, t2)
                self.add_bound(t2, t1)
            elif isinstance(t2, TypeVar):
                if not types.is_any(t1):
                    t2.find().upper_bound.add(t1)
                if not types.is_null(t1):
                    t2.find().lower_bound.add(t1)
            else:
                raise ValueError(f"Cannot equate nodes {t1} and {t2}")
        elif isinstance(t1, TypeVar):
            if isinstance(t2, TypeVar):
                t1.union(t2)
            elif isinstance(t2, ir.Type):
                if not types.is_any(t2):
                    t1.find().upper_bound.add(t2)
                if not types.is_null(t2):
                    t1.find().lower_bound.add(t2)
            else:
                raise ValueError(f"Cannot equate nodes {t1} and {t2}")
        else:
            raise ValueError(f"Cannot equate nodes {t1} and {t2}")

    def _collapse_node_supertype_cycles(self):
        # Create a directed graph from the node_super_types relationships
        G = nx.DiGraph()

        for v in self._tyvars():
            v = v.find()
            for w in v.node_super_types:
                w = w.find()
                G.add_edge(v, w)

        # Equate all type vars in the same SCC.
        for scc in nx.strongly_connected_components(G):
            if len(scc) > 1:
                cycle = list(scc)
                for t in cycle[1:]:
                    cycle[0].union(t)

    def dump(self):
        for v in self._tyvars():
            if v != v.find():
                print(f"{v} == {v.find()}")

        seen = set()
        for v in self._tyvars():
            v = v.find()
            if v.node.id in seen:
                continue
            seen.add(v.node.id)
            for w in v.node_super_types:
                print(f"{v} --> {w}")

        for v in self._tyvars():
            v = v.find()
            if v.lower_bound:
                print(f"{' | '.join(ir.value_to_string(t) for t in v.lower_bound)} <: {v}")
            if v.upper_bound:
                print(f"{v} <: {' | '.join(ir.value_to_string(t) for t in v.upper_bound)}")

    def type_bounds_compatible(self, tv1: TypeVar, tv2: TypeVar) -> bool:
        # Check that two type variables have compatible bounds.
        for lb in tv1.lower_bound:
            for ub in tv2.upper_bound:
                if not types.is_subtype(lb, ub):
                    return False
        for lb in tv2.lower_bound:
            for ub in tv1.upper_bound:
                if not types.is_subtype(lb, ub):
                    return False
        # TODO: not sure if this is needed. Leave out for now.
        # Check that all upper bounds are either value types or entity types, not both.
        if False:
            if all(types.is_value_type(ub) for ub in tv1.upper_bound) != all(types.is_value_type(ub) for ub in tv2.upper_bound):
                return False
            if all(not types.is_value_type(ub) for ub in tv1.upper_bound) != all(not types.is_value_type(ub) for ub in tv2.upper_bound):
                return False
        return True

    def compute_type_from_bounds(self, node: PyUnion[ir.Var, ir.Field], bound: OrderedSet[ir.Type]) -> Optional[ir.Type]:
        # Union all the types in the upper bound and intersect with the lower bound.
        # If we have {Any} in the union, change all the types to Set types.
        # Discard Any from the union.
        assert all(isinstance(t, (ir.ScalarType, ir.TupleType)) for t in bound), f"Bound {bound} contains non-scalar or non-tuple types."
        not_any = [t for t in bound if not types.is_any(t)]
        if not not_any:
            return None
        union = types.union(*not_any)
        return union

    def compute_type(self, tv: TypeVar, node: PyUnion[ir.Var, ir.Field], index: int, parent: ir.Node) -> ir.Type:
        root = tv.find()

        # Union all the types in the upper bound and intersect with the lower bound.
        # If we have {Any} in the union, change all the types to Set types.
        # Discard Any from the union.
        upper = self.compute_type_from_bounds(node, root.upper_bound)

        if upper is None:
            # We don't have any constraints that bound the type from above.
            if self.strict:
                self._complain(node, f"Could not infer a type for {ir.node_to_string(node).strip()}")
            return cast(ir.Type, types.Any)

        for lower in root.lower_bound:
            # TODO
            # Disable the check that the lower bound is satisfied.
            # If we don't declare all the subtype relationships, we'll end up with errors here.
            # Strict models should check this.
            # if not self.strict:
            #     continue
            if not types.is_subtype(lower, upper):
                name = TypeVar.pretty_name(node, index, parent)
                self._complain(node, f"{name} must be a supertype of {ir.type_to_string(lower)}, but has type {ir.type_to_string(upper)} instead.")

        return cast(ir.Type, upper)

    def _propagate_bounds(self):
        """Propagate bounds along node_super_types edges until a fixpoint is reached."""
        worklist = ordered_set(*self._tyvars())
        while worklist:
            sub = worklist.pop()
            sub = sub.find()
            for sup in sub.node_super_types:
                sup = sup.find()
                # Propagate upper bounds downward from supertype to subtype.
                n = len(sub.upper_bound)
                sub.upper_bound |= sup.upper_bound
                if n != len(sub.upper_bound):
                    worklist.add(sub)
                # Propagate lower bounds upward from subtype to supertype.
                n = len(sup.lower_bound)
                sup.lower_bound |= sub.lower_bound
                if n != len(sup.lower_bound):
                    worklist.add(sup)

    def _tyvars(self) -> Iterable[TypeVar]:
        """Return all the type variables in the graph."""
        return self.type_vars.values()

    def _equivalent_tyvars(self, tv: TypeVar) -> Iterable[TypeVar]:
        """Return the set of type variables that are equivalent to the given type variable."""
        tv = tv.find()
        return [v for v in self._tyvars() if v.find() == tv]

    def _unify_overloaded_relations(self):
        """Unify relations with the same name and arity that have compatible types."""
        worklist = ordered_set(*self._tyvars())
        while worklist:
            v = worklist.pop()
            v = v.find()

            # Non-strict models:
            # If there is no bound for the field, equate the relation with other relations of the same name
            # that have compatible types.
            # Then add back to the worklist all type variables in the relations.
            if not v.upper_bound:
                fields = []
                # If a tyvar has no upper bound, find all fields in the same equivalence class and
                # equate their enclosing relations.

                # 1. Get all the fields in the equivalence class.
                for x in self._equivalent_tyvars(v):
                    if isinstance(x.node, ir.Field):
                        fields.append(x.node)

                # 2. Get their enclosing relations.
                enclosing_relations = ordered_set()
                for f in fields:
                    # First find the enclosing relation of this field.
                    for r in self.model.relations:
                        for f2 in r.fields:
                            if f2.id == f.id:
                                enclosing_relations.add(r)
                                break

                # 3. Equate the relations with the same name and arity that have compatible types.
                changed = False
                for enclosing_relation in enclosing_relations:
                    for r in self.model.relations:
                        if r.id == enclosing_relation.id:
                            continue
                        if r.name == enclosing_relation.name and len(r.fields) == len(enclosing_relation.fields):
                            # Check for compatible types.
                            all_compat = True
                            for (f1, f2) in zip(r.fields, enclosing_relation.fields):
                                tv1 = self.get_type_var(f1).find()
                                tv2 = self.get_type_var(f2).find()
                                if self.type_bounds_compatible(tv1, tv2):
                                    all_compat = False
                                    break
                            if all_compat:
                                for (f1, f2) in zip(r.fields, enclosing_relation.fields):
                                    tv1 = self.get_type_var(f1).find()
                                    tv2 = self.get_type_var(f2).find()
                                    self.add_equality(tv1, tv2)
                                    assert tv1.find() == tv2.find()
                                    tv = tv1.find()
                                    # Enqueue all tyvars in the equivalence class of the fields.
                                    for x in self._equivalent_tyvars(tv):
                                        worklist.add(x)
                                    changed = True
                if changed:
                    # TODO This should be done incrementally on the new equivalents relations in the worklist.
                    self._collapse_node_supertype_cycles()
                    # There's no need to propagate bounds again since no changes
                    # are made to node_super_types.

    def solve(self):
        """Solve the type constraints."""

        if self.verbosity:
            print("\n")
            if self.verbosity > 1:
                ir.dump(self.model)
            else:
                print(self.model)

            print("\n")
            print("Constraints before solving:")
            self.dump()

        # Collapse all the node_super_types cycles into equalities.
        self._collapse_node_supertype_cycles()
        # Propagate bounds along node_super_types edges.
        self._propagate_bounds()
        # Unify overloaded relations.
        self._unify_overloaded_relations()

        if self.verbosity:
            print("\n")
            print("Constraints after solving:")
            self.dump()


@dataclass
class SubstituteTypes(visitor.DeprecatedPass):
    """
    A visitor that substitutes types back into the model.
    """
    env: TypeEnv = field(init=True)
    strict: bool = field(init=True)

    def handle_var(self, node: ir.Var, parent: ir.Node) -> ir.Var:
        x = self.env.get_type_var(node)
        t = self.env.compute_type(x, node, x.index, parent)
        if t is not None and not types.is_null(t):
            if isinstance(t, ir.UnionType):
                for t2 in t.types:
                    if not isinstance(t2, ir.ScalarType):
                        self.env._complain(node, f"Variable {node.name} inferred to be a non-scalar type {t2}. Variables must have scalar type.")
            elif not isinstance(t, ir.ScalarType):
                self.env._complain(node, f"Variable {node.name} inferred to be a non-scalar type {t}. Variables must have scalar type.")
            # Substitute the intersection of the inferred type with the declared type.
            new_type = SubstituteTypes._lower_type(node.type, t)
            return ir.Var(new_type, node.name)
        else:
            if self.strict:
                self.env._complain(node, f"Could not infer a type for variable {node.name}")
            return node

    def handle_field(self, node: ir.Field, parent: ir.Node) -> ir.Field:
        # Substitute the intersection of the inferred type with the declared type.
        changed = False
        if isinstance(node.type, ir.TupleType):
            new_types = []
            for i in range(len(node.type.types)):
                name = TypeVar.pretty_name(node, i, parent)
                x = self.env.get_type_var(node, i)
                t = self.env.compute_type(x, node, i, parent)
                if t is not None and not types.is_null(t):
                    if isinstance(t, ir.UnionType):
                        for t2 in t.types:
                            if not isinstance(t2, ir.ScalarType):
                                self.env._complain(node, f"{name} is inferred to be a non-scalar type {t2}. Variables must have scalar type.")
                    elif not isinstance(t, ir.ScalarType):
                        self.env._complain(node, f"{name} is inferred to be a non-scalar type {t}. Variables must have scalar type.")
                    new_type = SubstituteTypes._lower_type(node.type.types[i], t)
                    changed = True
                else:
                    if self.strict:
                        lower_name = name[0].lower() + name[1:]
                        self.env._complain(node, f"Could not infer a type for {lower_name}")
                    new_type = node.type.types[i]
                new_types.append(new_type)
            new_type = ir.TupleType(tuple(new_types))
        else:
            name = TypeVar.pretty_name(node, -1, parent)
            x = self.env.get_type_var(node)
            t = self.env.compute_type(x, node, x.index, parent)
            if t is not None and not types.is_null(t):
                if isinstance(t, ir.UnionType):
                    for t2 in t.types:
                        if not isinstance(t2, ir.ScalarType):
                            self.env._complain(node, f"{name} inferred to be a non-scalar type {t2}. Variables must have scalar type.")
                elif not isinstance(t, ir.ScalarType):
                    self.env._complain(node, f"{name} inferred to be a non-scalar type {t}. Variables must have scalar type.")
                new_type = SubstituteTypes._lower_type(node.type, t)
                changed = True
            else:
                if self.strict:
                    lower_name = name[0].lower() + name[1:]
                    self.env._complain(node, f"Could not infer a type for {lower_name}")
                new_type = node.type
        if changed:
            return ir.Field(node.name, new_type, node.input)
        else:
            return node

    # Intersect the declared type with the inferred type, special casing set types.
    @staticmethod
    def _lower_type(declared: ir.Type, inferred: ir.Type) -> ir.Type:
        if isinstance(declared, ir.SetType):
            if isinstance(inferred, ir.ScalarType):
                return ir.SetType(SubstituteTypes._lower_type(declared.element_type, inferred))
            elif isinstance(inferred, ir.UnionType):
                return types.union(*[SubstituteTypes._lower_type(declared, t) for t in inferred.types])
            else:
                raise ValueError(f"Set types must have scalar element types, but {inferred} is not a scalar type.")
        else:
            return types.intersect(declared, inferred)

@dataclass
class CollectTypeConstraints(visitor.DAGVisitor):
    """
    A visitor that collects type constraints on a model.
    """
    def __init__(self, env: TypeEnv):
        super().__init__()
        self.env = env

    def visit_model(self, node: ir.Model, parent: Optional[ir.Node]=None):
        # To avoid visiting overloads, add the overloads to the `seen` set
        # of DAGVisitor. This will trick the visitor into skipping them.
        for c in node.relations:
            for o in c.overloads:
                self.seen.add(o)
        super().visit_model(node, parent)

    def visit_scalartype(self, node: ir.Type, parent: Optional[ir.Node]=None):
        # Do not recurse.
        pass

    def visit_listtype(self, node: ir.ListType, parent: Optional[ir.Node]=None):
        # Do not recurse.
        pass

    def visit_settype(self, node: ir.SetType, parent: Optional[ir.Node]=None):
        # Do not recurse.
        pass

    def visit_uniontype(self, node: ir.UnionType, parent: Optional[ir.Node]=None):
        # Do not recurse.
        pass

    def visit_field(self, node: ir.Field, parent: Optional[ir.Node]=None):
        # Do not constrain field types to be <: Any.
        if isinstance(node.type, ir.SetType):
            if not types.is_any(node.type.element_type):
                self.env.add_bound(node, node.type.element_type)
        elif isinstance(node.type, ir.TupleType):
            for i in range(len(node.type.types)):
                x = self.env.get_type_var(node, i)
                if not types.is_any(node.type.types[i]):
                    self.env.add_bound(x, node.type.types[i])
        elif isinstance(node.type, ir.UnionType):
            for t in node.type.types:
                if not types.is_any(t):
                    self.env.add_bound(node, t)
        elif isinstance(node.type, ir.ScalarType):
            if not types.is_any(node.type):
                self.env.add_bound(node, node.type)
        # Do not recurse. No need to visit the type.
        pass

    def visit_var(self, node: ir.Var, parent: Optional[ir.Node]=None):
        # Do not constrain field types to be <: Any.
        if isinstance(node.type, ir.UnionType):
            for t in node.type.types:
                if not types.is_any(t):
                    self.env.add_bound(node, t)
        elif isinstance(node.type, ir.ScalarType):
            if not types.is_any(node.type):
                self.env.add_bound(node, node.type)
        # Do not recurse. No need to visit the type.
        pass

    def visit_default(self, node: ir.Default, parent: Optional[ir.Node]=None):
        # The variable's type should be a supertype of the default value.
        self.env.add_bound(node.value, node.var)
        # Recurse to add the constraints on the variable.
        return super().visit_default(node, parent)

    def visit_literal(self, node: ir.Literal, parent: Optional[ir.Node]=None):
        # Do not recurse. No need to visit the type.
        pass

    def visit_loop(self, node: ir.Loop, parent: Optional[ir.Node]=None):
        # The iterator should be a number.
        self.env.add_bound(node.iter, types.Number)
        return super().visit_loop(node, parent)

    def visit_update(self, node: ir.Update, parent: Optional[ir.Node]=None):
        if len(node.args) == len(node.relation.fields):
            # We have update R(x,y,z) where R is declared (t,u,v)
            # Bound each arg by the declared type of the field.
            for f, arg in zip(node.relation.fields, node.args):
                if f.input:
                    # Flow from argument to input field.
                    self.env.add_bound(arg, f)
                else:
                    # Flow from argument to field, and back.
                    self.env.add_equality(arg, f)
        return super().visit_update(node, parent)

    def visit_annotation(self, node: ir.Annotation, parent: Optional[ir.Node]=None):
        # Do not recurse. No need to visit the relation again.
        pass

    def visit_lookup(self, node: ir.Lookup, parent: Optional[ir.Node]=None):
        if len(node.args) == len(node.relation.fields):
            # We have lookup R(x,y,z) where R is declared (t,u,v)
            # Bound each arg by the declared type of the field.
            for f, arg in zip(node.relation.fields, node.args):
                if f.input:
                    # Flow from argument to input field.
                    self.env.add_bound(arg, f)
                else:
                    # Flow from argument to field, and back.
                    self.env.add_equality(arg, f)
        return super().visit_lookup(node, parent)

    def visit_aggregate(self, node: ir.Aggregate, parent: Optional[ir.Node]=None):
        agg = node.aggregation

        if len(agg.fields) < 3:
            return

        projection = agg.fields[0]
        group = agg.fields[1]

        inputs = []
        outputs = []
        for f in agg.fields[2:]:
            if f.input:
                inputs.append(f)
            else:
                outputs.append(f)

        # Now let's wire up the types.

        # Projection field.
        if isinstance(projection.type, ir.ScalarType):
            # In this case, we know the arity is 1, we just need to constrain the projection variables to be the same as the projection field types.
            if len(node.projection) == 1:
                self.env.add_equality(projection.type, node.projection[0])
        elif isinstance(projection.type, ir.TupleType):
            # In this case, we know the arity, we just need to constrain the projection variables to be the same as the projection field types.
            if len(node.projection) == len(projection.type.types):
                for i in range(len(node.projection)):
                    f = self.env.get_type_var(projection, i)
                    a = node.projection[i]
                    self.env.add_equality(f, a)

        # Group field.
        if isinstance(group.type, ir.ScalarType):
            # In this case, we know the arity is 1, we just need to constrain the group variables to be the same as the group field types.
            if len(node.group) == 1:
                self.env.add_equality(group.type, node.group[0])
        elif isinstance(group.type, ir.TupleType):
            # In this case, we know the arity, we just need to constrain the group variables to be the same as the group field types.
            if len(node.group) == len(group.type.types):
                for i in range(len(node.group)):
                    f = self.env.get_type_var(group, i)
                    a = node.group[i]
                    self.env.add_equality(f, a)

        # Inputs and outputs.
        if len(node.args) == len(inputs) + len(outputs):
            for a, f in zip(node.args[0:len(inputs)], inputs):
                # Flow from argument to input field.
                self.env.add_bound(a, f)

            for a, f in zip(node.args[len(inputs):], outputs):
                # Flow from argument to field, and back.
                self.env.add_equality(a, f)

        return super().visit_aggregate(node, parent)


@dataclass
class Typer(compiler.Pass):
    """
    A pass that performs type inference on a model.
    The pass also checks that the model is well-formed.
    Diagnostics are reported for ill-formed or ill-typed models.

    The main idea is to traverse the model and collect type constraints.
    These are then solved and substituted back into the model.
    """

    # Should we perform stricter checks on the inferred types?
    strict: bool = field(default=False, init=False)

    # How verbose to be with debug output, 0 is off.
    verbosity: int = field(default=0, init=False)

    def rewrite(self, model: ir.Model, cache) -> ir.Model:
        if self.verbosity:
            print("\n")
            print("\nInitial model:")
            if self.verbosity > 1:
                ir.dump(model)
            else:
                print(model)

        # Type inference.
        env = TypeEnv(model, self.strict, self.verbosity)
        collector = CollectTypeConstraints(env)
        model.accept(collector)

        if self.verbosity:
            print("Constraints:")
            env.dump()

        env.solve()

        # Substitute the types back into the model.
        subst = SubstituteTypes(env, self.strict)
        model2 = subst.walk(model)

        # Assert that there are no type errors
        if env.diags:
            error_count = len(env.diags)
            error_header = "TYPE ERROR\n" if error_count == 1 else f"{error_count} TYPE ERRORS\n"
            formatted_errors = [error_header] + [f"* (Node id={env.diags[i].node.id}) {env.diags[i].msg}" for i in range(error_count)] + ["-----"] + [str(model)]
            raise Exception("\n".join(formatted_errors))

        if self.verbosity:
            print("After substitution:")
            if self.verbosity > 1:
                ir.dump(model)
            else:
                print(model2)

        return model2
